# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__version__ = '1.0'

import abc
import logging
import os
import time
import warnings

from pyjob.exception import PyJobError, PyJobTaskLockedError
from pyjob.script import ScriptCollector, is_valid_script_path

ABC = abc.ABCMeta('ABC', (object, ), {})
logger = logging.getLogger(__name__)


class Task(ABC):
    """Abstract base class for executable tasks"""

    def __init__(self, script, *args, **kwargs):
        """Instantiate a new :obj:`~pyjob.task.Task`

        Parameters
        ----------
        script : :obj:`~pyjob.script.ScriptCollector`, :obj:`~pyjob.script.Script`, str, list, tuple
           A :obj:`str`, :obj:`list` or :obj:`tuple` of one or more script paths

        """
        self.pid = None
        self.locked = False
        if isinstance(script, ScriptCollector):
            self.script_collector = script
        else:
            self.script_collector = ScriptCollector(script)
        # These arguments are universal to all Task entities
        self.directory = os.path.abspath(kwargs.get('directory', '.'))
        self.nprocesses = kwargs.get('processes', 1)

    def __del__(self):
        """Exit function at instance deletion"""
        if not self.locked:
            self.lock()
        self.close()

    def __enter__(self):
        """Contextmanager entry function

        Note
        ----
        For further details see `PEP 343 <https://www.python.org/dev/peps/pep-0343/>`_.

        """
        return self

    def __exit__(self, *exc):
        """Contextmanager exit function

        Note
        ----
        For further details see `PEP 343 <https://www.python.org/dev/peps/pep-0343/>`_.

        """
        if not self.locked:
            self.lock()
        self.close()

    def __repr__(self):
        """Representation of the :obj:`~pyjob.task.Task`"""
        return '{}(pid={})'.format(self.__class__.__name__, self.pid)

    # ------------------ Abstract methods and properties ------------------

    @abc.abstractproperty
    def info(self):    # pragma: no cover
        """Abstract property to provide info about the :obj:`~pyjob.task.Task`"""
        pass

    @abc.abstractmethod
    def close(self):    # pragma: no cover
        """Abstract method to end :obj:`~pyjob.task.Task`"""
        pass

    @abc.abstractmethod
    def kill(self):    # pragma: no cover
        """Abstract method to forcefully terminate :obj:`~pyjob.task.Task`"""
        pass

    @abc.abstractmethod
    def _run(self):    # pragma: no cover
        """Abstract property to start execution of the :obj:`~pyjob.task.Task`"""
        pass

    # ------------------ Other task-specific general methods ------------------

    @property
    def completed(self):
        """Boolean to indicate :obj:`~pyjob.task.Task` completion"""
        return self.locked and not bool(self.info)

    @property
    def log(self):
        """The log file path"""
        return [script.log for script in self.script_collector]

    @property
    def script(self):
        """The script file path"""
        return [script.path for script in self.script_collector]

    def add_script(self, script):
        """Add further scripts to this :obj:`~pyjob.task.Task`

        Parameters
        ----------
        script : :obj:`~pyjob.script.Script`, str, list, tuple
           Something representing one or more scripts

        """
        if self.locked:
            raise PyJobTaskLockedError('This task is locked!')
        self.script_collector.add(script)

    def lock(self):
        """Lock this :obj:`~pyjob.task.Task`"""
        self.locked = True
        logger.debug('Locked %s [%d]', self.__class__.__name__, self.pid)

    def run(self):
        """Start the execution of this :obj:`~pyjob.task.Task`

        Raises
        ------
        :exc:`~pyjob.exception.PyJobError`
           One or more executable scripts required prior to execution
        :exc:`~pyjob.exception.PyJobTaskLockedError`
           Locked task, cannot restart or rerun

        """
        if self.locked:
            raise PyJobTaskLockedError('This task is locked!')
        if len(self.script_collector) < 1:
            raise PyJobError('One or more executable scripts required prior to execution')
        self.script_collector.dump()
        self._run()
        logger.debug('Started execution of %s [%d]', self.__class__.__name__, self.pid)
        self.lock()

    def wait(self, interval=30, monitor_f=None, success_f=None, check_success=None, monitor=None):
        """Method to wait for the completion of the current :obj:`~pyjob.task.Task`

        Parameters
        ----------
        interval : int, optional
           The interval to wait between checking (in seconds)
        monitor_f : func, optional
           A :obj:`callable` that is regularly invoked
        success_f : func, optional
           A :obj:`callable` to check for early termination of :obj:`~pyjob.task.Task`

        Note
        ----
        The `success_f` argument needs to accept a log file as input and return
        a :obj:`bool`.

        """
        if check_success:
            warnings.warn('This keyword argument has been deprecated, use success_f instead', DeprecationWarning)
            success_f = check_success

        if monitor:
            warnings.warn('This keyword argument has been deprecated, use monitor_f instead', DeprecationWarning)
            monitor_f = monitor

        def callable_checker(f):
            return bool(f and callable(f))

        do_check_success = callable_checker(success_f)
        if do_check_success:
            msg = 'Checking for %s %d success with function %s'
            logger.debug(msg, self.__class__.__name__, self.pid, success_f.__name__)
        do_monitor = callable_checker(monitor_f)
        while not self.completed:
            if do_check_success:
                for log in self.log:
                    if os.path.isfile(log) and success_f(log):
                        logger.debug("%s %d succeeded, run log: %s", self.__class__.__name__, self.pid, log)
                        self.kill()
            if do_monitor:
                monitor_f()
            time.sleep(interval)


class ClusterTask(Task):
    """Abstract base class for executable cluster tasks"""

    def __init__(self, *args, **kwargs):
        """Instantiate a new :obj:`~pyjob.task.ClusterTask`"""
        super(ClusterTask, self).__init__(*args, **kwargs)
        self.dependency = kwargs.get('dependency', [])
        self.max_array_size = kwargs.get('max_array_size', len(self.script))
        self.priority = kwargs.get('priority', None)
        self.queue = kwargs.get('queue', None)
        self.runtime = kwargs.get('runtime', None)
        self.shell = kwargs.get('shell', None)
        self.name = kwargs.get('name', 'pyjob')
        self.extra = kwargs.get('extra', [])

    @abc.abstractmethod
    def _create_runscript(self):
        """Utility method to create a :obj:`~pyjob.task.ClusterTask` runscript"""
        pass

    def get_array_bash_extension(self, jobsf, offset):
        """Get the array job bash extension for the ``runscript``

        Parameters
        ----------
        jobsf : str
           The file containing all scripts on a per-line basis
        offset : int
           The offset to be applied to the ``JOB_ARRAY_INDEX``

        Returns
        -------
        list
           A list of lines to be written to the ``runscript``

        Raises
        ------
        :exc:`ValueError`
           Invalid offset
        :exc:`ValueError`
           Valid job file required

        """
        if jobsf is None or not os.path.isfile(jobsf):
            raise ValueError('Valid job file required')
        if offset < 0:
            raise ValueError('Invalid offset')
        if offset > 0:
            script_def = 'script=$(awk "NR==$(({} + {}))" {})'.format(self.__class__.JOB_ARRAY_INDEX, offset, jobsf)
        else:
            script_def = 'script=$(awk "NR=={}" {})'.format(self.__class__.JOB_ARRAY_INDEX, jobsf)
        return [script_def, 'log=$(echo $script | sed "s/\\.${script##*.}/\\.log/")', '$script > $log 2>&1']



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__version__ = '1.0'

import logging
import re
import time
import uuid

from pyjob.cexec import cexec
from pyjob.exception import PyJobExecutableNotFoundError
from pyjob.script import Script
from pyjob.task import ClusterTask

logger = logging.getLogger(__name__)

RE_LINE_SPLIT_1 = re.compile(":\\s+")
RE_LINE_SPLIT_2 = re.compile("\\s+=\\s+")


class PortableBatchSystemTask(ClusterTask):
    """PortableBatchSystem executable :obj:`~pyjob.task.Task`"""

    JOB_ARRAY_INDEX = '$PBS_ARRAYID'
    SCRIPT_DIRECTIVE = '#PBS'

    @property
    def info(self):
        """:obj:`~pyjob.pbs.PortableBatchSystemTask` information"""
        if self.pid is None:
            return {}
        try:
            stdout = cexec(['qstat', '-f', str(self.pid)], permit_nonzero=True)
        except PyJobExecutableNotFoundError:
            return {}
        all_lines = stdout.splitlines()
        data = {}
        key, job_id = RE_LINE_SPLIT_1.split(all_lines[0], 1)
        data[key] = job_id
        for line in all_lines[1:]:
            line = line.strip()
            if 'Unknown queue destination' in line:
                return data
            else:
                kv = RE_LINE_SPLIT_2.split(line, 1)
                if len(kv) == 2:
                    data[kv[0]] = kv[1]
        return data

    def close(self):
        """Close this :obj:`~pyjob.pbs.PortableBatchSystemTask` after completion"""
        self.wait()

    def kill(self):
        """Immediately terminate the :obj:`~pyjob.pbs.PortableBatchSystemTask`"""
        if self.pid is None:
            return
        cexec(['qdel', str(self.pid)])
        logger.debug("Terminated task: %d", self.pid)

    def _run(self):
        """Method to initialise :obj:`~pyjob.pbs.PortableBatchSystemTask` execution"""
        runscript = self._create_runscript()
        runscript.write()
        self.pid = cexec(['qsub', runscript.path], directory=self.directory)
        logger.debug('%s [%d] submission script is %s', self.__class__.__name__, self.pid, runscript.path)

    def _create_runscript(self):
        """Utility method to create runscript"""
        runscript = Script(directory=self.directory, prefix='pbs_', suffix='.script', stem=str(uuid.uuid1().int))
        runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' -V')
        runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' -N {}'.format(self.name))
        if self.directory:
            cmd = '-w {}'.format(self.directory)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.priority:
            cmd = '-p {}'.format(self.priority)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.queue:
            cmd = '-q {}'.format(self.queue)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.runtime:
            h, m = divmod(self.runtime, 60)
            m, s = divmod(m, 60)
            cmd = '-l walltime={}:{}:{}'.format(h, m, s)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.shell:
            cmd = '-S {}'.format(self.shell)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.nprocesses:
            cmd = '-n {}'.format(self.nprocesses)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.extra:
            cmd = ' '.join(map(str, self.extra))
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if len(self.script) > 1:
            logf = runscript.path.replace('.script', '.log')
            jobsf = runscript.path.replace('.script', '.jobs')
            with open(jobsf, 'w') as f_out:
                f_out.write('\n'.join(self.script))
            cmd = '-t {}-{}%{}'.format(1, len(self.script), self.max_array_size)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' -o {}'.format(logf))
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' -e {}'.format(logf))
            runscript.extend(self.get_array_bash_extension(jobsf, 0))
        else:
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' -o {}'.format(self.log[0]))
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' -e {}'.format(self.log[0]))
            runscript.append(self.script[0])
        return runscript



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__version__ = '1.0'

import logging
import re
import uuid

from pyjob.cexec import cexec
from pyjob.exception import PyJobExecutableNotFoundError
from pyjob.script import Script
from pyjob.task import ClusterTask

logger = logging.getLogger(__name__)


class SlurmTask(ClusterTask):
    """Slurm executable :obj:`~pyjob.task.Task`"""

    JOB_ARRAY_INDEX = '$SLURM_ARRAY_TASK_ID'
    SCRIPT_DIRECTIVE = '#SBATCH'

    @property
    def info(self):
        """:obj:`~pyjob.slurm.SlurmTask` information"""
        if self.pid is None:
            return {}
        try:
            cexec(['squeue', '-j', str(self.pid)])
        except (PyJobExecutableNotFoundError, Exception):
            return {}
        else:
            return {'job_number': self.pid, 'status': 'Running'}

    def close(self):
        """Close this :obj:`~pyjob.slurm.SlurmTask` after completion"""
        self.wait()

    def kill(self):
        """Immediately terminate the :obj:`~pyjob.slurm.SlurmTask`"""
        if self.pid is None:
            return
        cexec(['scancel', str(self.pid)])
        logger.debug("Terminated task: %d", self.pid)

    def _run(self):
        """Method to initialise :obj:`~pyjob.slurm.SlurmTask` execution"""
        runscript = self._create_runscript()
        runscript.write()
        stdout = cexec(['sbatch', runscript.path], directory=self.directory)
        self.pid = int(stdout.strip().split()[-1])
        logger.debug('%s [%d] submission script is %s', self.__class__.__name__, self.pid, runscript.path)

    def _create_runscript(self):
        """Utility method to create runscript"""
        runscript = Script(directory=self.directory, prefix='slurm_', suffix='.script', stem=str(uuid.uuid1().int))
        runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' --export=ALL')
        runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' --job-name={}'.format(self.name))
        if self.dependency:
            cmd = '--depend=afterok:{}'.format(':'.join(map(str, self.dependency)))
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.queue:
            cmd = '-p {}'.format(self.queue)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.nprocesses:
            cmd = '-n {}'.format(self.nprocesses)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.directory:
            cmd = '--workdir={}'.format(self.directory)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if self.extra:
            cmd = ' '.join(map(str, self.extra))
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
        if len(self.script) > 1:
            logf = runscript.path.replace('.script', '.log')
            jobsf = runscript.path.replace('.script', '.jobs')
            with open(jobsf, 'w') as f_out:
                f_out.write('\n'.join(self.script))
            cmd = '--array={}-{}%{}'.format(1, len(self.script), self.max_array_size)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' ' + cmd)
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' -o {}'.format(logf))
            runscript.extend(self.get_array_bash_extension(jobsf, 0))
        else:
            runscript.append(self.__class__.SCRIPT_DIRECTIVE + ' -o {}'.format(self.log[0]))
            runscript.append(self.script[0])
        return runscript



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__version__ = '1.0'

from pyjob.factory import TaskFactory
from pyjob.misc import deprecate


class Job(object):
    @deprecate(0.3, msg='Use Task/TaskFactory instead')
    def __init__(self, qtype):
        self.qtype = qtype
        self._task = None
        self._locked = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._task:
            self._task.close()

    def __del__(self):
        if self._task:
            self._task.close()

    @property
    def finished(self):
        return self._task.completed

    @property
    def pid(self):
        return self._task.pid

    @property
    def log(self):
        return self._task.log

    @property
    def script(self):
        return self._task.script

    def kill(self):
        self._task.kill()

    def submit(self, script, *args, **kwargs):
        if self._locked:
            return
        kwargs['processes'] = kwargs.get('nproc', None)
        self._task = TaskFactory(self.qtype, script, *args, **kwargs)
        self._task.run()
        self._locked = True

    def stat(self):
        return self._task.info

    def wait(self, *args, **kwargs):
        self._task.wait(*args, **kwargs)



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__contributors__ = ['Adam Simpkin']
__version__ = '1.0'

import enum
import os
import sys

from pyjob.cexec import is_exe
from pyjob.exception import PyJobError
from pyjob.pool import Pool


@enum.unique
class ScriptProperty(enum.Enum):
    """Enumeration for :obj:`~pyjob.script.Script`-specific properties"""
    # Tried to extend Enum but operation not allowed in Python3.7
    # https://docs.python.org/3/library/enum.html#restricted-subclassing-of-enumerations
    if sys.platform.startswith('win'):
        PERL = ('', '.pl')
        PYTHON = ('', '.py')
        SHELL = ('', '.bat')
    else:
        PERL = ('#!/usr/bin/env perl', '.pl')
        PYTHON = ('#!/usr/bin/env python', '.py')
        SHELL = ('#!/bin/bash', '.sh')

    def __init__(self, shebang, suffix):
        self.shebang = shebang
        self.suffix = suffix


if sys.platform.startswith('win'):
    EXE_EXT = '.exe'
else:
    EXE_EXT = ''

SCRIPT_HEADER, SCRIPT_EXT = (ScriptProperty.SHELL.shebang, ScriptProperty.SHELL.suffix)


class ScriptCollector(object):
    """A :obj:`~pyjob.script.ScriptCollector` to store executable :obj:`~pyjob.script.Script` instances

    Examples
    --------

    >>> from pyjob.script import ScriptCollector, Script
    >>> collector = ScriptCollector(None)
    >>> for _ in range(5):
    ...     collector.add(Script())

    """

    def __init__(self, scripts):
        """Instantiate a new :obj:`~pyjob.script.ScriptCollector`"""
        self._container = []
        self._save_script(scripts)

    def __iter__(self):
        """Iterator function"""
        for script in self._container:
            yield script

    def __len__(self):
        """Length function"""
        return len(self._container)

    def __repr__(self):
        """Representation function"""
        return '{}(nscripts={})'.format(self.__class__.__name__, len(self))

    @property
    def scripts(self):
        """The script file paths"""
        return self._container

    @scripts.setter
    def scripts(self, scripts):
        """The script file paths

        Parameters
        ----------
        script : :obj:`~pyjob.script.Script`, str, list, tuple
           Something representing one or more scripts

        Raises
        ------
        :exc:`~pyjob.exception.PyJobError`
           Script cannot be found or is not executable

        """
        self._container = []
        self._save_script(scripts)

    def add(self, scripts):
        """Add one or more script file paths

        Parameters
        ----------
        script : :obj:`~pyjob.script.Script`, str, list, tuple
           Something representing one or more scripts

        Raises
        ------
        :exc:`~pyjob.exception.PyJobError`
           Script cannot be found or is not executable

        """
        self._save_script(scripts)

    def dump(self):
        """Write all scripts to disk if not already done"""
        for script in self._container:
            if not os.path.isfile(script.path):
                script.write()

    def _save_script(self, script):
        """Helper function to assess/standardise executable input

        Parameters
        ----------
        script : :obj:`~pyjob.executable.Script`, str, list, tuple
           Something representing one or more executables

        Raises
        ------
        :exc:`~pyjob.exception.PyJobError`
           Unrecognised executable input

        """
        if isinstance(script, Script):
            self._container.append(script)
        elif isinstance(script, str):
            script = Script.read(script)
            self._container.append(script)
        elif isinstance(script, list) or isinstance(script, tuple):
            for s in script:
                if isinstance(s, Script):
                    self._container.append(s)
                elif isinstance(s, str):
                    self._container.append(Script.read(s))
                else:
                    raise PyJobError('Unrecognised executable input')
        elif script is None:
            pass
        else:
            raise PyJobError('Unrecognised executable input')


class Script(list):
    """Simple extension to :obj:`list` to hold the contents for an executable script

    Examples
    --------

    >>> from pyjob import Script
    >>> script = Script(directory='.', prefix='example', stem='', suffix='.sh')
    >>> script.append('sleep 5')
    >>> print(script)
    #!/bin/bash
    sleep 5

    """

    def __init__(self,
                 shebang=ScriptProperty.SHELL.shebang,
                 directory='.',
                 prefix='tmp',
                 stem='pyjob',
                 suffix=ScriptProperty.SHELL.suffix):
        """Instantiate a new :obj:`~pyjob.script.Script`

        Parameters
        ----------
        shebang : str, optional
           The Shebang line in the :obj:`~pyjob.script.Script`
        directory : str, optional
           The directory for :obj:`~pyjob.script.Script` storage
        prefix : str, optional
           The :obj:`~pyjob.script.Script` filename prefix
        stem : str, optional
           The :obj:`~pyjob.script.Script` filename middle
        suffix : str, optional
           The :obj:`~pyjob.script.Script` filename suffix

        """
        self.directory = directory
        self.prefix = prefix
        self.stem = stem
        self.suffix = suffix
        self.shebang = shebang

    def __add__(self, other):
        """Combine two :obj:`~pyjob.script.Script` instances"""
        if self.shebang != other.shebang:
            raise TypeError('Invalid shebang combination')
        if self.suffix != other.suffix:
            raise TypeError('Invalid suffix combination')
        script = Script(stem=self.stem + '-' + other.stem, shebang=self.shebang, suffix=self.suffix)
        script.content = [line for script in [self, other] for line in script]
        return script

    def __str__(self):
        """Content of :obj:`~pyjob.script.Script`"""
        content = self[:]
        if len(self.shebang) > 0:
            content.insert(0, self.shebang)
        return '\n'.join(map(str, content))

    @property
    def content(self):
        """Getter method for :attr:`~pyjob.script.Script` content"""
        return self

    @content.setter
    def content(self, content):
        """Setter method for :attr:`~pyjob.script.Script` content"""
        while len(self) > 0:
            self.pop()
        self.extend(content)

    @property
    def directory(self):
        """Getter method for :attr:`~pyjob.script.Script.directory`"""
        return self._directory

    @directory.setter
    def directory(self, directory):
        """Setter method for :attr:`~pyjob.script.Script.directory`"""
        self._directory = os.path.abspath(directory)

    @property
    def log(self):
        """Path to the log of the the :obj:`~pyjob.script.Script`"""
        return self.path.rsplit('.', 1)[0] + '.log'

    @property
    def path(self):
        """Path to the :obj:`~pyjob.script.Script`"""
        return os.path.join(self.directory, self.prefix + self.stem + self.suffix)

    @property
    def shebang(self):
        """:obj:`~pyjob.script.Script` shebang"""
        return self._shebang

    @shebang.setter
    def shebang(self, value):
        """:obj:`~pyjob.script.Script` shebang"""
        self._shebang = value

    @property
    def suffix(self):
        """:obj:`~pyjob.script.Script` file suffix"""
        return self._suffix

    @suffix.setter
    def suffix(self, value):
        """:obj:`~pyjob.script.Script` file suffix"""
        if value is None or len(value) < 1 or '.' not in value:
            raise ValueError('Script suffix required!')
        self._suffix = value

    def write(self):
        """Write the :obj:`~pyjob.script.Script` to :attr:`~pyjob.script.Script.path`"""
        fname = self.path
        with open(fname, 'w') as f_out:
            f_out.write(str(self))
        os.chmod(fname, 0o777)

    @staticmethod
    def read(path):
        """Read a script file to construct a :obj:`~pyjob.script.Script`

        Examples
        --------

        >>> from pyjob import read_script
        >>> script = read_script('./example.sh')
        >>> print(script)
        #!/bin/bash
        sleep 5

        Parameters
        ----------
        path : str
           The path to a script file

        Returns
        -------
        :obj:`~pyjob.script.Script`
           A :obj:`~pyjob.script.Script` instance

        """
        directory, fname = os.path.split(path)
        fname, ext = os.path.splitext(fname)
        script = Script(directory=directory, prefix='', stem=fname, suffix=ext)
        with open(path, 'r') as f:
            lines = f.read().splitlines()
        if len(lines) > 0 and lines[0][:2] == '#!':
            script.shebang = lines.pop(0)
        else:
            script.shebang = ''
        script.extend(lines)
        return script


class LocalScriptCreator(object):
    """A :obj:`~pyjob.script.ScriptCollector` to store executable :obj:`~pyjob.script.Script`
    instances created in parallel using an input ``func`` to create the scripts.

    Examples
    --------

    >>> from pyjob.script import LocalScriptCreator, Script
    >>> script_creator = LocalScriptCreator(func, iterable, processes)
    >>> collector = script_creator.collector

    """

    def __init__(self, func=None, iterable=None, processes=1):
        """Instantiate a new :obj:`~pyjob.script.LocalScriptCreator`

        Parameters
        ----------
        func : func
            function to create :obj:`~pyjob.script.Script` with custom command
        iterable : iterable
            iterable argument to input into func
        processes : int
            Number of processes to generate scripts with

        """
        self.func = func
        self.iterable = iterable
        self.processes = processes

    def __call__(self, i):
        return self.func(i)

    @property
    def collector(self):
        script_collector = ScriptCollector(None)
        with Pool(processes=self.processes) as pool:
            script_collector.add(pool.map(self, self.iterable))
        return script_collector


def is_valid_script_path(fname):
    """Validate a script path

    Parameters
    ----------
    fname : str
        The path to a script file

    Returns
    -------
    bool

    """
    return is_exe(fname)



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__contibutors__ = ['Jens Thomas']
__version__ = '1.0'

import logging
import os
import signal
import subprocess
import sys
import warnings

from pyjob.exception import PyJobExecutableNotFoundError, PyJobExecutionError
from pyjob.misc import decode

logger = logging.getLogger(__name__)


def _insert_or_ignore(d, k, v):
    if k not in d:
        d[k] = v


def is_exe(fpath):
    """Status to indicate if a file is an executable

    Parameters
    ----------
    fpath : str
       The path to the file to be tested

    Returns
    -------
    bool

    """
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def which(executable):
    """Python-based mirror of UNIX ``which`` command

    Parameters
    ----------
    executable : str
       The path or name for an executable

    Returns
    -------
    str
       The absolute path to the executable, or ``None`` if not found

    Credits
    -------
    https://stackoverflow.com/a/377028/3046533

    """
    fpath, fname = os.path.split(executable)
    if fpath:
        if is_exe(executable):
            return executable
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, executable)
            if is_exe(exe_file):
                return exe_file
    return None


def cexec(cmd, permit_nonzero=False, **kwargs):
    """Function to execute a command

    Parameters
    ----------
    cmd : list
       The command to call
    permit_nonzero : bool, optional
       Allow non-zero return codes [default: False]
    **kwargs : dict, option
       Any keyword arguments accepted by :obj:`~subprocess.Popen`

    Returns
    -------
    str
       The processes' standard out

    Raises
    ------
    :exc:`PyJobExecutableNotFoundError`
       Cannot find executable
    :exc:`PyJobExecutionError`
       Execution exited with non-zero return code

    """
    logger.debug("Executing '%s'", " ".join(cmd))

    if os.name == 'nt':
        _insert_or_ignore(kwargs, 'bufsize', 0)
        _insert_or_ignore(kwargs, 'shell', 'False')
    if 'directory' in kwargs:
        warnings.warn('directory keywoard has been deprecated, use cwd instead', DeprecationWarning)
        kwargs['cwd'] = kwargs['directory']
        kwargs.pop('directory')
    _insert_or_ignore(kwargs, 'cwd', os.getcwd())
    _insert_or_ignore(kwargs, 'stdout', subprocess.PIPE)
    _insert_or_ignore(kwargs, 'stderr', subprocess.STDOUT)

    stdinstr = kwargs.get('stdin', None)
    if stdinstr and isinstance(stdinstr, str):
        kwargs['stdin'] = subprocess.PIPE

    executable = which(cmd[0])
    if executable is None:
        warnings.warn('executable not in PATH. provide absolute path to executable in future', DeprecationWarning)
    #      raise PyJobExecutableNotFoundError('Cannot find executable: %s' % cmd[0])
    #  cmd[0] = executable

    try:
        p = subprocess.Popen(cmd, **kwargs)
        if stdinstr:
            stdinstr = stdinstr.encode()
        stdout, stderr = p.communicate(input=stdinstr)
    except (KeyboardInterrupt, SystemExit):
        os.kill(p.pid, signal.SIGTERM)
        sys.exit(signal.SIGTERM)
    else:
        if stdout:
            stdout = decode(stdout).strip()
        if p.returncode == 0:
            return stdout
        elif permit_nonzero:
            logger.debug("Ignoring non-zero returncode %d for '%s'", p.returncode, " ".join(cmd))
            return stdout
        else:
            msg = "Execution of '{}' exited with non-zero return code ({})"
            raise PyJobExecutionError(msg.format(' '.join(cmd), p.returncode))



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__version__ = '1.0'

import chardet.universaldetector
import os
import sys
import tempfile
import warnings

from pyjob.exception import PyJobError


def decode(byte_s):
    """Decode a string by guessing the encoding

    Parameters
    ----------
    byte_s : bytes
       The :obj:`bytes` to decode

    Returns
    -------
    :obj:`str`
       `byte_s` decoded

    Raises
    ------
    :exc:`PyJobError`
       Unable to infer string encoding

    """
    detector = chardet.universaldetector.UniversalDetector()
    for line in byte_s.splitlines():
        detector.feed(line)
        if detector.done:
            break
    detector.close()
    if detector.result['confidence'] < 0.98:
        raise PyJobError('Unable to infer string encoding')
    return byte_s.decode(detector.result['encoding'])


def deprecate(version, msg=None):
    """Decorator to deprecate Python classes and functions

    Parameters
    ----------
    version : str
       A string containing the version with which the callable is removed
    msg : str, optional
       An additional message that will be displayed alongside the default message


    Examples
    --------
    Enable :obj:`~DeprecationWarning` messages to be displayed.

    >>> import warnings
    >>> warnings.simplefilter('default')

    Decorate a simple Python function without additional message

    >>> @deprecate('0.0.0')
    ... def sum(a, b):
    ...     return a + b
    >>> sum(1, 2)
    deprecated.py:34: DeprecationWarning: sum has been deprecated and will be removed in version 0.0.0!
      warnings.warn(message, DeprecationWarning)
    3

    Decorate a simple Python function with additional message

    >>> @deprecate('0.0.1', msg='Use XXX instead!')
    ... def sum(a, b):
    ...     return a + b
    >>> sum(2, 2)
    deprecated.py:34: DeprecationWarning: sum has been deprecated and will be removed in version 0.0.0!
      warnings.warn(message, DeprecationWarning)
    4

    Decorate an entire Python class

    >>> @deprecate('0.0.2')
    ... class Obj(object):
    ...     pass
    >>> Obj()
    deprecated.py:34: DeprecationWarning: Obj has been deprecated and will be removed in version 0.0.2!
      warnings.warn(message, DeprecationWarning)
    <__main__.Obj object at 0x7f8ee0f1ead0>

    Decorate a Python class method

    >>> class Obj(object):
    ...     def __init__(self, v):
    ...         self.v = v
    ...     @deprecate('0.0.3')
    ...     def mul(self, other):
    ...         return self.v * other.v
    >>> Obj(2).mul(Obj(3))
    deprecated.py:34: DeprecationWarning: mul has been deprecated and will be removed in version 0.0.3!
      warnings.warn(message, DeprecationWarning)
    6

    Decorate a Python class staticmethod

    >>> class Obj(object):
    ...     @staticmethod
    ...     @deprecate('0.0.4')
    ...     def sub(a, b):
    ...         return a - b
    ...
    >>> Obj.sub(2, 1)
    deprecated.py:34: DeprecationWarning: sub has been deprecated and will be removed in version 0.0.4!
      warnings.warn(message, DeprecationWarning)
    1

    Decorate a Python class classmethod

    >>> class Obj(object):
    ...     CONST = 5
    ...     @classmethod
    ...     @deprecate('0.0.5')
    ...     def sub(cls, a):
    ...         return a - cls.CONST
    ...
    >>> Obj().sub(5)
    deprecated.py:34: DeprecationWarning: sub has been deprecated and will be removed in version 0.0.5!
      warnings.warn(message, DeprecationWarning)
    0

    """

    def deprecate_decorator(callable_):
        def warn(*args, **kwargs):
            message = "%s has been deprecated and will be removed in version %s!" % (callable_.__name__, version)
            if msg:
                message += " - %s" % msg
            warnings.warn(message, DeprecationWarning)
            return callable_(*args, **kwargs)

        return warn

    return deprecate_decorator


@deprecate(0.3, msg='use pyjob.stopwatch.StopWatch')
def is_script(f):
    from pyjob.script import is_valid_script_path
    return is_valid_script_path(f)


@deprecate(0.3, msg='use pyjob.stopwatch.StopWatch')
def StopWatch():
    from pyjob.stopwatch import StopWatch
    return StopWatch()


@deprecate(0.3, msg='use pyjob.script.Script')
def make_script(cmd, **kwargs):
    from pyjob.script import Script
    script = Script(**kwargs)
    if isinstance(cmd, list) and isinstance(cmd[0], list) \
            or isinstance(cmd, tuple) and isinstance(cmd[0], tuple) \
            or isinstance(cmd, list) and isinstance(cmd[0], tuple) \
            or isinstance(cmd, tuple) and isinstance(cmd[0], list):
        for c in cmd:
            script.append(' '.join(map(str, c)))
    elif isinstance(cmd, list) or isinstance(cmd, tuple):
        script.append(' '.join(map(str, cmd)))
    script.write()
    return script.path


@deprecate(0.3, msg='use pyjob.script.Script')
def make_python_script(cmd, directory=None, prefix="tmp", stem=None):
    return make_script(
        cmd, directory=directory, shebang="#!/usr/bin/env python", prefix=prefix, stem=stem, suffix=".py")


@deprecate(0.3)
def tmp_dir(directory=None, prefix="tmp", suffix=""):
    """Return a filename for a temporary directory

    Parameters
    ----------
    directory : str, optional
       Path to a directory to write the files to.
    prefix : str, optional
       A prefix to the temporary filename
    suffix : str, optional
       A suffix to the temporary filename

    """
    return tempfile.mkdtemp(dir=directory, prefix=prefix, suffix=suffix)


@deprecate(0.3)
def tmp_file(delete=False, directory=None, prefix="tmp", stem=None, suffix=""):
    """Return a filename for a temporary file

    The naming convention of scripts will be ``prefix`` + ``stem`` + ``suffix``.

    Parameters
    ----------
    delete : bool, optional
       Delete the file, thus return name only [default: True]
    directory : str, optional
       Path to a directory to write the files to
    prefix : str, optional
       A prefix to the temporary filename
    stem : str, optional
       The steam part of the script name
    suffix : str, optional
       A suffix to the temporary filename

    """
    if directory is None:
        directory = tempfile.gettempdir()
    if stem is None:
        tmpf = tempfile.NamedTemporaryFile(delete=delete, dir=directory, prefix=prefix, suffix=suffix)
        tmpf.close()
        return tmpf.name
    else:
        tmpf = os.path.join(directory, "".join([prefix, stem, suffix]))
        if not delete:
            open(tmpf, 'w').close()
        return tmpf



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__date__ = '29 Jul 2018'
__version__ = '1.0'

import multiprocessing.pool
import sys


class Pool(multiprocessing.pool.Pool):
    """:obj:`~multiprocessing.pool.Pool` of processes to allow concurrent method calls

    Examples
    --------

    >>> from pyjob import Pool
    >>> with Pool(processes=2) as pool:
    ...     pool.map(<func>, <iterable>)

    """
    if sys.version_info.major < 3:

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.terminate()



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'

import argparse
import logging
import os
import sys

from pyjob import TaskFactory, __version__
from pyjob.factory import TASK_PLATFORMS


def main():
    p = argparse.ArgumentParser()
    p.add_argument('-d', '--directory', default='.')
    p.add_argument('-p', '--platform', choices=TASK_PLATFORMS.keys(), default='local')
    p.add_argument('-t', '--threads', default=1, type=int, dest='processes')
    p.add_argument('--chdir', default=False, action='store_true')
    p.add_argument('--permit-nonzero', default=False, action='store_true')
    p.add_argument('--verbose', action='count')
    p.add_argument('--version', action='version', version='pyjob ' + __version__)
    p.add_argument('executables', nargs='+')

    kwargs = vars(p.parse_args())
    platform = kwargs.pop('platform')
    executables = [os.path.abspath(f) for f in kwargs.pop('executables')]
    kwargs['directory'] = os.path.abspath(kwargs['directory'])

    verbosity_lvl = kwargs.pop('verbose')
    if verbosity_lvl == 0:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.DEBUG)

    with TaskFactory(platform, executables, **kwargs) as task:
        task.run()

    sys.exit(0)


if __name__ == '__main__':
    main()



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__version__ = '1.0'

import importlib
import logging

from pyjob.exception import PyJobUnknownTaskPlatform

TASK_PLATFORMS = {
    'local': ('pyjob.local', 'LocalTask'),
    'lsf': ('pyjob.lsf', 'LoadSharingFacilityTask'),
    'pbs': ('pyjob.pbs', 'PortableBatchSystemTask'),
    'slurm': ('pyjob.slurm', 'SlurmTask'),
    'sge': ('pyjob.sge', 'SunGridEngineTask'),
    'torque': ('pyjob.torque', 'TorqueTask'),
}

logger = logging.getLogger(__name__)


def TaskFactory(platform, *args, **kwargs):
    """Accessibility function for any :obj:`~pyjob.task.Task`

    Examples
    --------

    >>> from pyjob import TaskFactory
    >>> with TaskFactory('local', script) as task:
    ...     task.run()

    Parameters
    ----------
    platform : str
       The platform to create the :obj:`~pyjob.task.Task` on
    *args : tuple
       Any positional arguments relevant to the :obj:`~pyjob.task.Task`
    **kwargs : dict
       Any keyword arguments relevant to the :obj:`~pyjob.task.Task`

    Raises
    ------
    :exc:`~pyjob.exception.PyJobUnknownTaskPlatform`
       Unknown platform

    """
    platform = platform.lower()
    if platform in TASK_PLATFORMS:
        logger.debug('Found requested platform in available task list')
        module, class_ = TASK_PLATFORMS[platform]
        return getattr(importlib.import_module(module), class_)(*args, **kwargs)
    else:
        raise PyJobUnknownTaskPlatform('Unknown platform: %s' % platform)



# MIT License
#
# Copyright (c) 2017-18 Felix Simkovic
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = 'Felix Simkovic'
__version__ = '1.0'

import logging
import multiprocessing
import os
import sys
import time
import uuid

from pyjob.cexec import cexec
from pyjob.task import Task

CPU_COUNT = multiprocessing.cpu_count()

logger = logging.getLogger(__name__)


class LocalTask(Task):
    """Locally executable :obj:`~pyjob.task.Task`"""

    def __init__(self, *args, **kwargs):
        """Instantiate a new :obj:`~pyjob.local.LocalTask`"""
        super(LocalTask, self).__init__(*args, **kwargs)
        self.queue = multiprocessing.Queue()
        self.kill_switch = multiprocessing.Event()
        self.processes = []
        self.chdir = kwargs.get('chdir', False)
        self.permit_nonzero = kwargs.get('permit_nonzero', False)
        self._killed = False

    @property
    def nprocesses(self):
        """Getter for the number of concurrent :obj:`~pyjob.local.LocalProcess`"""
        return self._nprocesses

    @nprocesses.setter
    def nprocesses(self, nprocesses):
        """Setter for the number of concurrent :obj:`~pyjob.local.LocalProcess`"""
        if nprocesses > CPU_COUNT:
            logger.warning('More processes requested than available CPUs')
        self._nprocesses = nprocesses

    @property
    def info(self):
        """:obj:`~pyjob.local.LocalTask` information"""
        if any(proc.is_alive() for proc in self.processes):
            return {'job_number': self.pid, 'status': 'Running'}
        return {}

    def close(self):
        """Close this :obj:`~pyjob.local.LocalTask` after completion"""
        if self._killed:
            return
        for proc in self.processes:
            proc.join()
        self.kill()

    def kill(self):
        """Immediately terminate the :obj:`~pyjob.local.LocalTask`"""
        if self._killed:
            return
        if not self.kill_switch.is_set():
            self.kill_switch.set()
        # This is a requirement to avoid access to memory-inaccessible processes
        # The queue gets flushed by triggering the kill_switch
        for proc in self.processes:
            proc.join()
        for proc in self.processes:
            proc.terminate()
        logger.debug("Terminated task: %d", self.pid)
        self._killed = True

    def _run(self):
        """Method to initialise :obj:`~pyjob.local.LocalTask` execution"""
        if self._killed:
            return
        for _ in range(self.nprocesses):
            proc = LocalProcess(
                self.queue,
                self.kill_switch,
                directory=self.directory,
                chdir=self.chdir,
                permit_nonzero=self.permit_nonzero)
            proc.start()
            self.processes.append(proc)
        for script in self.script:
            self.queue.put(script)
        for _ in self.processes:
            self.queue.put(None)
        self.queue.close()
        self.pid = uuid.uuid1().int
        time.sleep(0.1)


class LocalProcess(multiprocessing.Process):
    """Extension to :obj:`multiprocessing.Process` for :obj:`~pyjob.local.LocalTask`"""

    def __init__(self, queue, kill_switch, directory=None, permit_nonzero=False, chdir=False):
        """Instantiate a :obj:`~pyjob.local.LocalProcess`

        Parameters
        ----------
        queue : :obj:`~multiprocessing.Queue`
           An instance of a :obj:`~multiprocessing.Queue`
        kill_switch : obj
           An instance of a :obj:`~multiprocessing.Event`
        directory : str, optional
           The directory to execute the jobs in
        permit_nonzero : bool, optional
           Allow non-zero return codes

        Warning
        -------
        This object should only be instantiated by :obj:`~pyjob.local.LocalTask`!

        """
        super(LocalProcess, self).__init__()
        self.queue = queue
        self.kill_switch = kill_switch
        self.directory = directory
        self.permit_nonzero = permit_nonzero
        self.chdir = chdir

    def run(self):
        """Method representing the :obj:`~pyjob.local.LocalProcess` activity"""
        for job in iter(self.queue.get, None):
            if self.kill_switch.is_set():
                continue
            if self.chdir:
                directory = os.path.dirname(job)
            else:
                directory = self.directory
            log = os.path.splitext(job)[0] + '.log'
            with open(log, 'w') as f:
                cexec([job], cwd=directory, stdout=f, permit_nonzero=self.permit_nonzero)
